#!/usr/bin/env python

"""
Search Python abstract syntax trees for nodes of a particular type having a
particular textual value.

Copyright (C) 2008 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import compiler
import os

__version__ = "0.1"

def search_recursive(directory, term_type, term, op=None):

    """
    Search files within the filesystem below 'directory' for terms having the
    given 'term_type' whose value matches the specified 'term'.
    """

    results = []
    for path, directories, filenames in os.walk(directory):
        for filename in filenames:
            if os.path.splitext(filename)[-1] == os.path.extsep + "py":
                results += search_file(os.path.join(path, filename), term_type, term, op)
    return results

def search_file(filename, term_type, term, op=None):

    """
    Search the file with the given 'filename' for terms having the given
    'term_type' whose value matches the specified 'term'.
    """

    node = compiler.parseFile(filename)
    cls = getattr(compiler.ast, term_type)
    return search_tree(node, cls, term, op, filename)

def search_tree(node, cls, term, op=None, filename=None):

    """
    Search the tree rooted at the given 'node' for nodes of the given class
    'cls' for content matching the specified 'term'.

    Return a list of results of the form (node, value, filename).
    """

    results = []

    if isinstance(node, cls):
        if op is None:
            results.append((node, None, filename))
        else:
            for child in node.getChildren():
                if isinstance(child, (str, unicode, int, float, long, bool)) and op(unicode(child)):
                    results.append((node, child, filename))
                    break

    # Search within nodes, even if matches have already been found.

    for child in node.getChildNodes():
        results += search_tree(child, cls, term, op, filename)

    return results

def expand_results(results):

    """
    Expand the given 'results', making a list containing tuples of the form
    (filename, line number, line, value).
    """

    expanded = []

    for node, value, filename in results:
        if filename is not None:
            line = linecache.getline(filename, node.lineno).rstrip()
        else:
            line = None

        expanded.append((filename, node.lineno, line, value))

    return expanded

# Command syntax.

syntax_description = """
    [ -n | --line-number ]
    [ -p | --print-token ]
    ( ( -t TERM_TYPE ) | ( --type=TERM_TYPE ) )
    [ ( -e PATTERN ) | ( --regexp=PATTERN ) ]
    ( ( ( -r | -R | --recursive ) DIRECTORY ) | FILENAME )
    """

# Main program.

if __name__ == "__main__":
    import sys
    import cmdsyntax
    import re
    import linecache

    # Match command arguments.

    syntax = cmdsyntax.Syntax(syntax_description)
    syntax_matches = syntax.get_args(sys.argv[1:])

    try:
        args = syntax_matches[0]
    except IndexError:
        print "Syntax:"
        print syntax_description
        sys.exit(1)

    # Get the search details.

    term_type = args["TERM_TYPE"]
    term = args.get("PATTERN")

    if term is None:
        op = None
    else:
        op = re.compile(term).match

    # Perform the search either in a single file or in a directory hierarchy.

    if args.has_key("FILENAME"):
        results = search_file(args["FILENAME"], term_type, term, op)
    else:
        results = search_recursive(args["DIRECTORY"], term_type, term, op)

    # Present the results.

    for filename, lineno, line, value in expand_results(results):
        format = "%s:"
        output = [filename]

        if args.has_key("n") or args.has_key("line-number"):
            format += "%d:"
            output.append(lineno)

        if args.has_key("p"):
            if value is not None:
                format += "%r:"
                output.append(value)
            else:
                format += "%s:"
                output.append("<%s>" % term_type)

        format += " %s"
        output.append(line)

        print format % tuple(output)

# vim: tabstop=4 expandtab shiftwidth=4
